"""
Overview:
    Simulation for cli entry.
"""
import io
import traceback
from contextlib import contextmanager
from functools import partial
from typing import Optional, List, Callable, Mapping, ContextManager
from unittest.mock import patch

from ..capture import capture_output, capture_exit

__all__ = [
    'simulate_entry', 'EntryRunResult',
]


class EntryRunResult:
    """
    Overview:
        Run result of one entry.
    """

    def __init__(self, exitcode: int, stdout: Optional[str], stderr: Optional[str], error: Optional[BaseException]):
        """
        Constructor of :class:`EntryRunResult`.

        :param exitcode: Exit code.
        :param stdout: Output in standard output stream.
        :param stderr: Output in standard error stream.
        :param error: Uncaught exception raised inside.
        """
        self.__exitcode = exitcode
        self.__stdout = stdout
        self.__stderr = stderr
        self.__error = error

    @property
    def exitcode(self) -> int:
        """
        Exit code.
        """
        return self.__exitcode

    @property
    def stdout(self) -> Optional[str]:
        """
        Output in standard output stream.
        """
        return self.__stdout

    @property
    def stderr(self) -> Optional[str]:
        """
        Output in standard error stream.
        """
        return self.__stderr

    @property
    def error(self) -> Optional[BaseException]:
        """
        Uncaught exception raised inside.
        """
        return self.__error

    def _assert_okay_message(self) -> str:
        with io.StringIO() as sf:
            pp = partial(print, file=sf)
            if self.error is not None:
                pp(f'Exitcode - {self.exitcode!r}, with uncaught exception:')
                traceback.print_tb(self.error.__traceback__, file=sf)
                pp(f'{type(self.error)}: {self.error.args!r}')
            else:
                pp(f'Exitcode - {self.exitcode!r}.')

            if self.stdout:
                pp(f'---------------------------------')
                pp(f'[Stdout]')
                pp(self.stdout)
                pp()

            if self.stderr:
                pp(f'---------------------------------')
                pp(f'[Stderr]')
                pp(self.stderr)
                pp()

            return sf.getvalue()

    def assert_okay(self):
        assert self.exitcode == 0 and self.error is None, self._assert_okay_message()


# See: https://stackoverflow.com/questions/36136480/what-is-pythons-default-exit-code
_OKAY_EXITCODE = 0x0
_ERROR_EXITCODE = 0x1
_USAGE_EXITCODE = 0x2


@contextmanager
def _mock_argv(argv: Optional[List[str]] = None) -> ContextManager:
    if argv is not None:
        with patch('sys.argv', argv):
            yield
    else:
        yield


@contextmanager
def _mock_environ(envs: Optional[Mapping[str, str]] = None) -> ContextManager:
    if envs is not None:
        with patch.dict('os.environ', envs, clear=False):
            yield
    else:
        yield


def simulate_entry(entry: Callable, argv: Optional[List[str]] = None,
                   envs: Optional[Mapping[str, str]] = None) -> EntryRunResult:
    """
    Overview:
        CLI entry's simulation.

    :param entry: Entry function, should be a simple function without any arguments.
    :param argv: Command line arguments. Default is ``None``, which means do not mock ``sys.argv``.
    :param envs: Environment arguments. Default is ``None``, which means do not mock ``os.environ``.
    :returns: A result object, in form of :class:`EntryRunResult`.

    Examples::
        We create a simple CLI code with `click package <https://click.palletsprojects.com/>`_, \
            named ``test_cli1.py``

        .. code-block:: python
           :linenos:

            import sys
            import click

            @click.command('cli1', help='CLI-1 example')
            @click.option('-c', type=int, help='optional C value', default=None)
            @click.argument('a', type=int)
            @click.argument('b', type=int)
            def cli1(a, b, c):
                if c is None:
                    print(f'{a} + {b} = {a + b}')
                else:
                    print(f'{a} + {b} + {c} = {a + b + c}', file=sys.stderr)

            if __name__ == '__main__':
                cli1()

        When we can try to simulate it.

            >>> from hbutils.testing import simulate_entry
            >>> from test_cli1 import cli1
            >>> r1 = simulate_entry(cli1, ['cli1', '2', '3'])
            >>> print(r1.exitcode)
            0
            >>> print(r1.stdout)
            2 + 3 = 5

            >>> r2 = simulate_entry(cli1, ['cli1', '2', '3', '-c', '24'])  # option
            >>> print(r2.exitcode)
            0
            >>> print(r2.stderr)
            2 + 3 + 24 = 29

            >>> r3 = simulate_entry(cli1, ['cli', '--help'])  # help
            >>> print(r3.stdout)
            Usage: cli [OPTIONS] A B
              CLI-1 example
            Options:
              -c INTEGER  optional C value
              --help      Show this message and exit.

            >>> r4 = simulate_entry(cli1, ['cli', 'dklsfj'])  # misusage
            >>> print(r4.exitcode)
            2
            >>> print(r4.stderr)
            Usage: cli [OPTIONS] A B
            Try 'cli --help' for help.
            Error: Invalid value for 'A': 'dklsfj' is not a valid integer.

        .. note::
            Please note that if there is uncaught exception raised inside the entry function, \
                it will be caught and put into ``error`` property instead of be being printed \
                to ``stderr``. For example

            >>> from hbutils.testing import simulate_entry
            >>> def my_cli():
            ...     raise ValueError(233)
            >>>
            >>> r = simulate_entry(my_cli)
            >>> print(r.exitcode)  # will be 0x1
            1
            >>> print(r.stdout)  # nothing
            >>> print(r.stderr)  # nothing as well
            >>> print(repr(r.error))  # HERE!!!
            ValueError(233)
    """
    try:
        with capture_output() as _out, capture_exit(_OKAY_EXITCODE) as _exit, \
                _mock_argv(argv), _mock_environ(envs):
            entry()
    except BaseException as err:
        return EntryRunResult(_ERROR_EXITCODE, _out.stdout, _out.stderr, err)
    else:
        return EntryRunResult(_exit.exitcode, _out.stdout, _out.stderr, None)



import ctypes
from typing import BinaryIO

from .base import CFixedType

__all__ = [
    'CBoolType',
    'c_bool',
]


class CBoolType(CFixedType):
    """
    Overview:
        Boolean type.
    """

    def __init__(self, size: int):
        """
        Constructor of :class:`CBoolType`.

        :param size: Size of boolean type.
        """
        CFixedType.__init__(self, size)
        self.__size = size

    def read(self, file: BinaryIO) -> bool:
        """
        Read boolean value.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: Boolean value.
        """
        return any(file.read(self.__size))

    def write(self, file: BinaryIO, val: bool):
        """
        Write boolean value to binary IO object.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: Boolean value to write.
        """
        file.write(b'\x00' * (self.__size - 1) + (b'\x01' if val else b'\x00'))


c_bool = CBoolType(ctypes.sizeof(ctypes.c_bool))
"""
Overview:
    Reading and writing bool value like C language.

Examples::
    >>> import io
    >>> from hbutils.binary import c_bool
    >>> 
    >>> with io.BytesIO(b'\\x01\\x00\\x01\\x00') as file:
    ...     print(c_bool.read(file))
    ...     print(c_bool.read(file))
    ...     print(c_bool.read(file))
    ...     print(c_bool.read(file))
    True
    False
    True
    False
    >>> with io.BytesIO() as file:
    ...     c_bool.write(file, True)
    ...     c_bool.write(file, False)
    ...     c_bool.write(file, True)
    ...     c_bool.write(file, False)
    ...     print(file.getvalue())
    ... 
    b'\\x01\\x00\\x01\\x00'
"""



from typing import Dict

from ..os import is_windows

__all__ = [
    'hostfile', 'get_hosts', 'get_localhost_ip',
]


def hostfile() -> str:
    """
    Overview:
        Get host file in this os.

    :return: Host file path in this os.

    .. note::
        This should be ``/etc/hosts`` on Linux and macOS, but ``c:\windows\system32\drivers\etc\hosts`` on Windows.

    """
    return r"c:\windows\system32\drivers\etc\hosts" if is_windows() else '/etc/hosts'


def get_hosts() -> Dict[str, str]:
    """
    Overview:
        Get hosts content in form of dictionary.

    :return: A dictionary of the hosts file.

    Examples::
        >>> from hbutils.system import get_hosts
        >>> get_hosts()
        {'hansbug-VirtualBox': '127.0.0.1', 'localhost': '127.0.0.1'}

    """
    resdict = {}
    with open(hostfile(), 'r') as hf:
        for line in hf:
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            _content, *_ = line.split('#', maxsplit=1)
            ip_address, *hostnames = _content.strip().split()
            for host in hostnames:
                resdict[host.strip()] = ip_address.strip()

    return resdict


_LOCALHOST = 'localhost'
_DEFAULT_LOCALHOST = '127.0.0.1'


def get_localhost_ip() -> str:
    """
    Overview:
        Get ip address of localhost.
        It will be checked in hosts, and just return ``127.0.0.1`` is not found.

    :return: IP address of ``localhost``.

    Examples::
        >>> from hbutils.system import get_localhost_ip
        >>> get_localhost_ip()
        '127.0.0.1'

    """
    return get_hosts().get(_LOCALHOST, _DEFAULT_LOCALHOST)



"""
Overview:
    Useful functions for build representation format of object.
"""
from typing import List, Tuple

__all__ = [
    'get_repr_info',
]


def get_repr_info(cls: type, args: List[Tuple]) -> str:
    """
    Overview:
        Get representation information for object.
        Can be used in ``__repr__`` method for class.

    Arguments:
        - cls (:obj:`type`): Object's type.
        - args (:obj:`List[Tuple]`): Argument display information.

    Returns:
        - repr (:obj:`str`): Representation string.

    Examples::
        >>> from hbutils.model import get_repr_info
        >>> class Sum:
        ...     def __init__(self, a, b):
        ...         self.__a = a
        ...         self.__b = b
        ...     def __repr__(self):
        ...         return get_repr_info(
        ...             cls=self.__class__,
        ...             args=[
        ...                 ('b', lambda: self.__b, lambda: self.__b is not None),
        ...                 ('a', lambda: self.__a),
        ...             ]
        ...         )
        ...
        >>> Sum(1, 2)
        <Sum b: 2, a: 1>
        >>> Sum(1, None)
        <Sum a: 1>
        >>> Sum(None, None)
        <Sum a: None>
    """
    _data_items = []
    for item in args:
        if isinstance(item, tuple):
            if len(item) == 2:
                name, fd = item
                if isinstance(fd, tuple):
                    _data_func, _present_func = fd
                else:
                    _data_func, _present_func = fd, lambda: True
            elif len(item) == 3:
                name, _data_func, _present_func = item
            else:
                raise ValueError('Tuple\'s length should be 2 or 3 but {actual} found.'.format(actual=repr(len(item))))

            if _present_func():
                _data_items.append('{name}: {data}'.format(name=name, data=_data_func()))
        else:
            raise TypeError(
                'Argument item should be tuple but {actual} found.'.format(actual=repr(type(item).__name__)))

    if _data_items:
        return '<{cls} {data}>'.format(cls=cls.__name__, data=', '.join(_data_items))
    else:
        return '<{cls}>'.format(cls=cls.__name__)



import re
from dataclasses import dataclass
from typing import List, Tuple, Optional
from urllib import parse as urlparse

__all__ = [
    'urlsplit', 'SplitURL',
]


def _split_params(params: str) -> List[Tuple[str, Optional[str]]]:
    name_value_pairs = re.split(r'[&;]', params)
    result = []
    for name_value_pair in name_value_pairs:
        # Split the pair string into a naive, encoded (name, value) pair.
        name_value = name_value_pair.split('=', 1)

        if len(name_value) == 1:
            # 'param' => ('param', None)
            name, value = name_value[0], None
        else:
            # 'param=value' => ('param', 'value')
            # 'param=' => ('param', '')
            name, value = name_value

        name = urlparse.unquote_plus(name)
        if value is not None:
            value = urlparse.unquote_plus(value)

        result.append((name, value))

    return result


def _split_path(path: str) -> List[str]:
    return [urlparse.unquote(segment) for segment in path.split('/')]


@dataclass
class SplitURL:
    """
    Overview:
        Splitted url object.
    """
    url: str
    scheme: str
    host: str
    path: str
    query: str
    fragment: str

    @property
    def query_dict(self):
        """
        Query dict.
        """
        retval = {}
        for key, value in _split_params(self.query or ''):
            if key in retval:
                if isinstance(retval[key], list):
                    retval[key].append(value)
                else:
                    retval[key] = [retval[key], value]
            else:
                retval[key] = value

        return retval

    @property
    def path_segments(self) -> List[str]:
        """
        Separated path segments.
        """
        return _split_path(self.path)

    @property
    def filename(self) -> Optional[str]:
        """
        Filename of current url, return ``None`` when path is empty.
        """
        return self.path_segments[-1]

    def __str__(self):
        """
        Original url.
        """
        return self.url

    def __repr__(self):
        """
        Repr format of :class:`SplitURL`.
        """
        content = ', '.join([f'{key}={value!r}' for key, value in [
            ('scheme', self.scheme),
            ('host', self.host),
            ('path', self.path),
            ('query', self.query_dict),
            ('fragment', self.fragment),
        ] if value])

        return f'{self.__class__.__name__}({content})'


def urlsplit(url: str) -> SplitURL:
    """
    Overview:
        Split url into 5 parts (scheme, host, path, query and fragment).

    :param url: Original url string.
    :return: :class:`SplitURL` object.

    Examples::
        >>> from hbutils.system import urlsplit
        >>>
        >>> sp = urlsplit('https://www.baidu.com/dslkjf/sdfhk/asdasd.png?q=1&v=kdjf&q=2#fff')
        >>> sp
        SplitURL(scheme='https', host='www.baidu.com', path='/dslkjf/sdfhk/asdasd.png', query={'q': ['1', '2'], 'v': 'kdjf'}, fragment='fff')
        >>> repr(sp)
        "SplitURL(scheme='https', host='www.baidu.com', path='/dslkjf/sdfhk/asdasd.png', query={'q': ['1', '2'], 'v': 'kdjf'}, fragment='fff')"
        >>>
        >>> sp.scheme
        'https'
        >>> sp.host
        'www.baidu.com'
        >>> sp.path
        '/dslkjf/sdfhk/asdasd.png'
        >>> sp.query
        'q=1&v=kdjf&q=2'
        >>> sp.fragment
        'fff'
        >>>
        >>> sp.query_dict
        {'q': ['1', '2'], 'v': 'kdjf'}
        >>> sp.path_segments
        ['', 'dslkjf', 'sdfhk', 'asdasd.png']
        >>> sp.filename
        'asdasd.png'

    """
    splitted = urlparse.urlsplit(url)
    return SplitURL(url, splitted.scheme, splitted.netloc, splitted.path, splitted.query, splitted.fragment)



"""
Overview:
    Final class implement (not a custom design pattern, but useful for designing).
"""

__all__ = ['FinalMeta']


class FinalMeta(type):
    """
    Overview:
        A meta class for making one class be final (unable to be extended by other classes).
    """

    def __new__(mcs, name, bases, attrs):
        """
        Overview:
            Creation process of new finalized class.
        Arguments:
            - name (:obj:`str`): Name of the new created class
            - bases (:obj:`Tuple[type]`): Base classes of the new created class
            - attrs (:obj: `Dict[str, Any]`): Attached attributes (such as method and fields) of the new created class
        Example:
            >>> class FinalClass(metaclass=FinalMeta):  # this is a final class
            >>>     pass
            >>>
            >>> class TryToExtendFinalClass(FinalMeta):  # TypeError will be raised in compile time
            >>>     pass
            >>>
        """
        for b in bases:
            if isinstance(b, FinalMeta):
                raise TypeError("Type {name} is a final class, which is not an acceptable common type."
                                .format(name=repr(b.__name__)))
        return type.__new__(mcs, name, bases, attrs)



"""
Overview:
    Backport support of :class:`tempfile.TemporaryDirectory` in python3.7 on Windows.
"""
import os
import platform
import shutil
import tempfile
import warnings
import weakref

__all__ = [
    'TemporaryDirectory',
]

_python_version_tuple = tuple(map(int, platform.python_version_tuple()))

if _python_version_tuple >= (3, 8):
    from tempfile import TemporaryDirectory
else:
    class TemporaryDirectory(object):
        """
        .. note::
            **This class is copied from python3.8's tempfile library.**

            Because PermissionError will be raised when use native TemporaryDirectory on **Windows python3.7**.
            This class should be removed when python3.7 is no longer supported.

            See `tempfile.TemporaryDirectory <https://docs.python.org/3/library/tempfile.html#tempfile.TemporaryDirectory>`_
            for more details.

        Create and return a temporary directory.  This has the same
        behavior as mkdtemp but can be used as a context manager.  For
        example:

            with TemporaryDirectory() as tmpdir:
                ...

        Upon exiting the context, the directory and everything contained
        in it are removed.
        """

        def __init__(self, suffix=None, prefix=None, dir=None):
            self.name = tempfile.mkdtemp(suffix, prefix, dir)
            self._finalizer = weakref.finalize(
                self, self._cleanup, self.name,
                warn_message="Implicitly cleaning up {!r}".format(self))

        @classmethod
        def _rmtree(cls, name):
            def onerror(func, path, exc_info):
                if issubclass(exc_info[0], PermissionError):
                    def resetperms(path):
                        try:
                            os.chflags(path, 0)
                        except AttributeError:
                            pass
                        os.chmod(path, 0o700)

                    try:
                        if path != name:
                            resetperms(os.path.dirname(path))
                        resetperms(path)

                        try:
                            os.unlink(path)
                        # PermissionError is raised on FreeBSD for directories
                        except (IsADirectoryError, PermissionError):
                            cls._rmtree(path)
                    except FileNotFoundError:
                        pass
                elif issubclass(exc_info[0], FileNotFoundError):
                    pass
                else:
                    raise

            shutil.rmtree(name, onerror=onerror)

        @classmethod
        def _cleanup(cls, name, warn_message):
            cls._rmtree(name)
            warnings.warn(warn_message, ResourceWarning)

        def __repr__(self):
            return "<{} {!r}>".format(self.__class__.__name__, self.name)

        def __enter__(self):
            return self.name

        def __exit__(self, exc, value, tb):
            self.cleanup()

        def cleanup(self):
            if self._finalizer.detach():
                self._rmtree(self.name)



from bisect import bisect_right
from typing import Sequence, Union, Tuple, Callable

__all__ = [
    'linear_map',
]


def linear_map(points: Union[Sequence[float], Sequence[Tuple[float, float]]]) -> Callable[[float], float]:
    """
    Overview:
        Multiple-staged linear gradient calculation based on float number.

    :param points: Points for this linear mapping. If the sequence consists of float numbers, it will be seen as the \
        simple linear mapping. If the elements are binary tuples (contains 2 float numbers), it means the x-range is \
        assigned.
    :return: A callable function for linear mapping.

    Examples::
        - Simple Linear Mapping

        >>> from hbutils.algorithm import linear_map
        >>>
        >>> f = linear_map((0, 1, 0.5))
        >>> f(0)
        0.0
        >>> f(0.25)
        0.5
        >>> f(1 / 3)
        0.6666666666666666
        >>> f(0.5)
        1.0
        >>> f(2 / 3)
        0.8333333333333334
        >>> f(0.75)
        0.75
        >>> f(1)
        0.5

        - Complex Linear Mapping (x values are customized)

        >>> f = linear_map(((-0.2, 0), (0.7, 1), (1.1, 0.5)))
        >>> f(-0.2)
        0.0
        >>> f(0)
        0.22222222222222227
        >>> f(0.25)
        0.5000000000000001
        >>> f(1 / 3)
        0.5925925925925927
        >>> f(0.5)
        0.7777777777777778
        >>> f(2 / 3)
        0.9629629629629631
        >>> f(0.7)
        1.0
        >>> f(0.75)
        0.9375
        >>> f(0.8)
        0.875
        >>> f(1)
        0.625
        >>> f(1.1)
        0.5

    """
    assert points, f'Points should be non-empty, but {points!r} found.'

    try:
        xys = [(x, y) for x, y in points]
    except TypeError:
        pts = list(points)
        n = len(pts)
        xys = [(i / (n - 1), y) for i, y in enumerate(points)]

    for i, (x1_, x2_) in enumerate(zip(xys[:-1], xys[1:])):
        assert x1_ < x2_, f'The former x value should be no more than the latter one, ' \
                          f'but {x1_} (at {i}) >= {x2_} (at {i + 1}) found.'

    def _linear(x: float) -> float:
        if xys[0][0] <= x <= xys[-1][0]:
            _index = bisect_right([x_ for x_, _ in xys[:-1]], x) - 1
            x1, y1 = xys[_index]
            x2, y2 = xys[_index + 1]

            r = (x - x1) / (x2 - x1)
            return (1 - r) * y1 + r * y2
        else:
            raise ValueError(f'Invalid x value, [{xys[0][0]!r}, {xys[-1][0]!r}] expected but {x!r} found.')

    return _linear



import hashlib

__all__ = [
    'md5', 'sha1',
    'sha224', 'sha256', 'sha384', 'sha512',
    'sha3',
]


def _hash_algorithm(algo, binary: bytes):
    p = algo()
    p.update(binary)
    return p.hexdigest()
    pass


def md5(binary: bytes) -> str:
    """
    Overview:
        MD5 hash.

    Arguments:
        - binary (:obj:`bytes`): Binary data to be hashed.

    Returns:
        - digest (:obj:`str`): MD5 digest string.

    Examples::
        >>> from hbutils.encoding import md5
        >>> md5(b'')
        'd41d8cd98f00b204e9800998ecf8427e'
        >>> md5(b'this is a word')
        'cdfc9527f76e296c76cdb331ac2d1d88'
    """
    return _hash_algorithm(hashlib.md5, binary)


def sha1(binary: bytes) -> str:
    """
    Overview:
        SHA1 hash.

    Arguments:
        - binary (:obj:`bytes`): Binary data to be hashed.

    Returns:
        - digest (:obj:`str`): SHA1 digest string.

    Examples::
        >>> from hbutils.encoding import sha1
        >>> sha1(b'')
        'da39a3ee5e6b4b0d3255bfef95601890afd80709'
        >>> sha1(b'this is a word')
        '7bf417a6503e185ea6352525b96a4d6ef3b9609b'
    """
    return _hash_algorithm(hashlib.sha1, binary)


def sha224(binary: bytes) -> str:
    """
    Overview:
        SHA224 hash.

    Arguments:
        - binary (:obj:`bytes`): Binary data to be hashed.

    Returns:
        - digest (:obj:`str`): SHA224 digest string.

    Examples::
        >>> from hbutils.encoding import sha224
        >>> sha224(b'')
        'd14a028c2a3a2bc9476102bb288234c415a2b01f828ea62ac5b3e42f'
        >>> sha224(b'this is a word')
        '7b994bcffbc9a3689941e541a2e639c20e321b763b744c0d84f6899c'
    """
    return _hash_algorithm(hashlib.sha224, binary)


def sha256(binary: bytes) -> str:
    """
    Overview:
        SHA256 hash.

    Arguments:
        - binary (:obj:`bytes`): Binary data to be hashed.

    Returns:
        - digest (:obj:`str`): SHA256 digest string.

    Examples::
        >>> from hbutils.encoding import sha256
        >>> sha256(b'')
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
        >>> sha256(b'this is a word')
        '91ccca153f5d3739af1f0d304d033f193b25208d44d371c1304877a6503471bf'
    """
    return _hash_algorithm(hashlib.sha256, binary)


def sha384(binary: bytes) -> str:
    """
    Overview:
        SHA384 hash.

    Arguments:
        - binary (:obj:`bytes`): Binary data to be hashed.

    Returns:
        - digest (:obj:`str`): SHA384 digest string.

    Examples::
        >>> from hbutils.encoding import sha384
        >>> sha384(b'')
        '38b060a751ac96384cd9327eb1b1e36a21fdb71114be07434c0cc7bf63f6e1da274edebfe76f65fbd51ad2f14898b95b'
        >>> sha384(b'this is a word')
        '37ae885a635d78e9067fdfe297e2d50d70de3670e4066bd541ac259b3ecdffb39a4c7b97211cb3967eac2188bb4b95f6'
    """
    return _hash_algorithm(hashlib.sha384, binary)


def sha512(binary: bytes) -> str:
    """
    Overview:
        SHA512 hash.

    Arguments:
        - binary (:obj:`bytes`): Binary data to be hashed.

    Returns:
        - digest (:obj:`str`): SHA512 digest string.

    Examples::
        >>> from hbutils.encoding import sha512
        >>> sha512(b'')
        'cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e'
        >>> sha512(b'this is a word')
        '6e80a2fd5dcf382b443ddaa619833208a4f4ac9bbedb760d6c8aac26a12da8dc6e2f6bc809091915db2e2d19bb66682db89e46c646a8ea7a5c6da57bf1a42b5c'
    """
    return _hash_algorithm(hashlib.sha512, binary)


def sha3(binary: bytes, n: int = 256) -> str:
    """
    Overview:
        SHA3 hash.

    Arguments:
        - binary (:obj:`bytes`): Binary data to be hashed.

    Returns:
        - digest (:obj:`str`): SHA3 digest string.
        - n (:obj:`int`): Length of SHA3, default is 256.

    Examples::
        >>> from hbutils.encoding import sha3
        >>> sha3(b'')
        'a7ffc6f8bf1ed76651c14756a061d662f580ff4de43b49fa82d80a4b80f8434a'
        >>> sha3(b'this is a word')
        '3e5a3507f625c8464fb282b451c8d82d78c3a232645e55c846d55551a7fe667c'
        >>> sha3(b'this is a word', n=224)
        'e0271d2734fc2c1a6dfcb6051bec6dc59e5f7fbec4b0d42ef1faee64'
    """
    return _hash_algorithm(getattr(hashlib, f'sha3_{n}'), binary)



"""
Overview:
    Function operations for nested structure.
"""
from ..reflection import dynamic_call

__all__ = [
    'nested_map'
]


def nested_map(f, s):
    """
    Overview:
        Map the nested structure with a function.

    Arguments:
        - f: The given function.
        - s: Nested structure.

    Returns:
        - result: Mapped nested structure.

    Examples::
        >>> from hbutils.collection import nested_map
        >>> nested_map(lambda x: x + 1, [
        ...     2, 3, (4, {'x': 2, 'y': 4}),
        ...     {'a': 3, 'b': (4, 5)},
        ... ])
        [3, 4, (5, {'x': 3, 'y': 5}), {'a': 4, 'b': (5, 6)}]
        >>> nested_map(lambda x, p: (x + 1) *  len(p), [
        ...     2, 3, (4, {'x': 2, 'y': 4}),
        ...     {'a': 3, 'b': (4, 5)},
        ... ])
        [3, 4, (10, {'x': 9, 'y': 15}), {'a': 8, 'b': (15, 18)}]
        >>> nested_map(lambda: 233, [
        ...     2, 3, (4, {'x': 2, 'y': 4}),
        ...     {'a': 3, 'b': (4, 5)},
        ... ])
        [233, 233, (233, {'x': 233, 'y': 233}), {'a': 233, 'b': (233, 233)}]
    """
    _df = dynamic_call(f)

    def _recursion(sval, p):
        if isinstance(sval, dict):
            return type(sval)({k: _recursion(v, (*p, k)) for k, v in sval.items()})
        elif isinstance(sval, (list, tuple)):
            return type(sval)(_recursion(v, (*p, i)) for i, v in enumerate(sval))
        else:
            return _df(sval, p)

    return _recursion(s, ())



import struct
from typing import BinaryIO


class CIOType:
    """
    Overview:
        Basic IO type.
        Used as base class of all the IO types.
    """

    def read(self, file: BinaryIO):
        """
        Read from binary IO object.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: Reading result.

        .. warning::
            Need to be implemented.
        """
        raise NotImplementedError  # pragma: no cover

    def write(self, file: BinaryIO, val):
        """
        Write object to binary IO object.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: Object to write.

        .. warning::
            Need to be implemented.
        """
        raise NotImplementedError  # pragma: no cover


class CFixedType(CIOType):
    """
    Overview:
        Type with fixed size (such as ``int``, ``uint`` and ``float``).
    """

    def __init__(self, size: int):
        """
        Constructor of :class:`CFixedType`.

        :param size: Size of the type.
        """
        self.__size = size

    @property
    def size(self) -> int:
        """
        Size of the given type.
        """
        return self.__size

    def read(self, file: BinaryIO):
        raise NotImplementedError  # pragma: no cover

    def write(self, file: BinaryIO, val):
        raise NotImplementedError  # pragma: no cover


class CRangedIntType(CFixedType):
    """
    Overview:
        Type with fixed size and range (such as ``int`` and ``uint``).
    """

    def __init__(self, size: int, minimum: int, maximum: int):
        """
        Constructor of :class:`CRangedIntType`.

        :param size: Size of the type.
        :param minimum: Min value of the type.
        :param maximum: Max value of the type.
        """
        CFixedType.__init__(self, size)
        self.__size = size
        self.__minimum = minimum
        self.__maximum = maximum

    @property
    def minimum(self) -> int:
        """
        Min value of the type.
        """
        return self.__minimum

    @property
    def maximum(self) -> int:
        """
        Max value of the type.
        """
        return self.__maximum

    def read(self, file: BinaryIO):
        raise NotImplementedError  # pragma: no cover

    def write(self, file: BinaryIO, val):
        raise NotImplementedError  # pragma: no cover


class CMarkedType(CFixedType):
    """
    Overview:
        Type with struct mark, which can be directly read by ``struct`` module.
    """

    def __init__(self, mark: str, size: int):
        """
        Constructor of :class:`CMarkedType`.

        :param mark: Mark of the type.
        :param size: Size of the type.
        """
        CFixedType.__init__(self, size)
        self.__mark = mark

    @property
    def mark(self) -> str:
        """
        Mark of the type, will be used to read from binary data with ``struct`` module.
        """
        return self.__mark

    def read(self, file: BinaryIO):
        """
        Read from binary with ``struct`` module.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: Result value.
        """
        r, = struct.unpack(self.mark, file.read(self.size))
        return r

    def write(self, file: BinaryIO, val):
        """
        Write value to binary IO with ``struct`` module.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: Writing value.
        """
        file.write(struct.pack(self.mark, float(val)))



import ctypes
from typing import Dict, BinaryIO, List

from .base import CRangedIntType
from .uint import CUnsignedIntType

__all__ = [
    'CSignedIntType',
    'c_int8', 'c_int16', 'c_int32', 'c_int64',
    'c_byte', 'c_short', 'c_int', 'c_long', 'c_longlong',
]


class CSignedIntType(CRangedIntType):
    """
    Overview:
        Signed int type.
    """

    def __init__(self, size: int):
        """
        Constructor of :class:`CSignedIntType`.

        :param size: Size of the type.
        """
        self.__size = size
        self._unit = CUnsignedIntType(size)
        self.__half = 1 << (8 * self.__size - 1)
        CRangedIntType.__init__(
            self, self.__size,
            self._unit.minimum - self.__half,
            self._unit.maximum - self.__half
        )

    def read(self, file: BinaryIO) -> int:
        """
        Read signed int value.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :return: Signed int value.
        """
        uval = self._unit.read(file)
        if uval < self.__half:
            return uval
        else:
            return uval - (self.__half << 1)

    def write(self, file: BinaryIO, val: int):
        """
        Write signed int value to binary IO object.

        :param file: Binary file, ``io.BytesIO`` is supported as well.
        :param val: Signed int value to write.
        """
        if not isinstance(val, int):
            raise TypeError(f'Int value expected, but {repr(val)} found.')
        elif not (self.minimum <= val <= self.maximum):
            raise ValueError(f'Signed int value within '
                             f'[{self.minimum}, {self.maximum}] expected, but {repr(val)} found.')

        fval = val if val >= 0 else val + (self.__half << 1)
        self._unit.write(file, fval)


c_int8 = CSignedIntType(ctypes.sizeof(ctypes.c_int8))
"""
Overview:
    Reading and writing signed integer with 8-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_int8
    >>> 
    >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef') as file:
    ...     print(c_int8.read(file))
    ...     print(c_int8.read(file))
    ...     print(c_int8.read(file))
    ...     print(c_int8.read(file))
    -34
    -83
    -66
    -17
    >>> with io.BytesIO() as file:
    ...     c_int8.write(file, -34)
    ...     c_int8.write(file, -83)
    ...     c_int8.write(file, -66)
    ...     c_int8.write(file, -17)
    ...     print(file.getvalue())
    b'\\xde\\xad\\xbe\\xef'
"""
c_int16 = CSignedIntType(ctypes.sizeof(ctypes.c_int16))
"""
Overview:
    Reading and writing signed integer with 16-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_int16
    >>> 
    >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef\\x12\\x34\\x56\\xf7') as file:
    ...     print(c_int16.read(file))
    ...     print(c_int16.read(file))
    ...     print(c_int16.read(file))
    ...     print(c_int16.read(file))
    -21026
    -4162
    13330
    -2218
    >>> with io.BytesIO() as file:
    ...     c_int16.write(file, -21026)
    ...     c_int16.write(file, -4162)
    ...     c_int16.write(file, 13330)
    ...     c_int16.write(file, -2218)
    ...     print(file.getvalue())
    b'\\xde\\xad\\xbe\\xef\\x124V\\xf7'
"""
c_int32 = CSignedIntType(ctypes.sizeof(ctypes.c_int32))
"""
Overview:
    Reading and writing signed integer with 32-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_int32
    >>> 
    >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef\\x12\\x34\\x56\\xf7') as file:
    ...     print(c_int32.read(file))
    ...     print(c_int32.read(file))
    -272716322
    -145345518
    >>> with io.BytesIO() as file:
    ...     c_int32.write(file, -272716322)
    ...     c_int32.write(file, -145345518)
    ...     print(file.getvalue())
    b'\\xde\\xad\\xbe\\xef\\x124V\\xf7'
"""
c_int64 = CSignedIntType(ctypes.sizeof(ctypes.c_int64))
"""
Overview:
    Reading and writing signed integer with 64-bits.
    
Examples::
    >>> import io
    >>> from hbutils.binary import c_int64
    >>> 
    >>> with io.BytesIO(b'\\xde\\xad\\xbe\\xef\\x12\\x34\\x56\\xf7') as file:
    ...     print(c_int64.read(file))
    -624254242407928354
    >>> with io.BytesIO() as file:
    ...     c_int64.write(file, -624254242407928354)
    ...     print(file.getvalue())
    b'\\xde\\xad\\xbe\\xef\\x124V\\xf7'
"""

_EXIST_TYPES: List[CSignedIntType] = [
    c_int8,
    c_int16,
    c_int32,
    c_int64,
]
_SIZE_TO_INT_TYPE: Dict[int, CSignedIntType] = {
    item.size: item
    for item in _EXIST_TYPES
}


def _get_from_raw(tp) -> CSignedIntType:
    return _SIZE_TO_INT_TYPE[ctypes.sizeof(tp)]


c_byte = _get_from_raw(ctypes.c_byte)
"""
Alias for :data:`c_uint8`.
"""
c_short = _get_from_raw(ctypes.c_short)
"""
Alias for :data:`c_int16`.
"""
c_int = _get_from_raw(ctypes.c_int)
"""
Alias for :data:`c_int32` (in 64-bits OS).

.. note::
    Size of :data:`c_int` is the same as that in C language, which mainly depends on CPU and OS.
"""
c_long = _get_from_raw(ctypes.c_long)
"""
Alias for :data:`c_int64` (in 64-bits OS).

.. note::
    Size of :data:`c_long` is the same as that in C language, which mainly depends on CPU and OS.
"""
c_longlong = _get_from_raw(ctypes.c_longlong)
"""
Alias for :data:`c_int64` (in 64-bits OS).

.. note::
    Size of :data:`c_longlong` is the same as that in C language, which mainly depends on CPU and OS.
"""



"""
Overview:
    Functions for directory processing.
"""
import errno
import os
import shutil

from .file import glob

__all__ = [
    'copy', 'remove',
    'getsize',
]


def _single_copy(src: str, dst: str):
    try:
        shutil.copytree(src, dst)  # copy directory
    except OSError as exc:
        if exc.errno in (errno.ENOTDIR, errno.EINVAL):
            shutil.copy(src, dst)  # copy file
        else:
            raise  # pragma: no cover


def copy(src1: str, src2: str, *srcn_dst: str):
    """
    Overview:
        Copy files or directories.

        No less than 2 arguments are accepted.
        When the last path is an existing path, all the fore paths will be copied to this path.
        Otherwise, the first path will be copied to the last path (exactly 2 arguments are accepted in this \
            case, or ``NotADirectoryError`` will be raised).

        From `Stack Overflow - Copy file or directories recursively in Python \
        <https://stackoverflow.com/a/1994840/6995899>`_.

    .. note::
        You can use this like ``cp -rf`` command on unix.

    Examples::
        >>> import os
        >>> from hbutils.system import copy
        >>>
        >>> os.listdir('.')
        ['test', 'README.md', 'cloc.sh', 'LICENSE', 'codecov.yml', 'pytest.ini', 'Makefile', 'setup.py', 'requirements-test.txt', 'requirements-doc.txt', 'requirements.txt']
        >>>
        >>> copy('cloc.sh', 'new_cloc.sh')  # copy file
        >>> copy('test', 'new_test')  # copy directory
        >>> os.listdir('.')
        ['test', 'README.md', 'cloc.sh', 'LICENSE', 'codecov.yml', 'new_test', 'pytest.ini', 'Makefile', 'setup.py', 'requirements-test.txt', 'requirements-doc.txt', 'requirements.txt', 'new_cloc.sh']
        >>>
        >>> os.makedirs('new_path_1')
        >>> copy('*.txt', 'new_path_1')  # copy to new path
        >>> os.listdir('new_path_1')
        ['requirements-test.txt', 'requirements-doc.txt', 'requirements.txt']
        >>>
        >>> os.makedirs('new_path_2')
        >>> copy('*.txt', 'test/system/**/*.py', 'new_path_2')  # copy plenty of files to new path
        >>> print(*os.listdir('new_path_2'), sep=\'\\n\')
        test_version.py
        test_file.py
        test_type.py
        test_package.py
        test_implementation.py
        requirements-test.txt
        __init__.py
        test_directory.py
        requirements-doc.txt
        requirements.txt
    """
    *srcs, dst = (src1, src2, *srcn_dst)
    if os.path.exists(dst) and os.path.isdir(dst):  # copy to directory
        for file in glob(*srcs):
            _, name = os.path.split(file)
            _single_copy(file, os.path.join(dst, name))

    else:  # copy to file
        if len(srcs) > 1:
            raise NotADirectoryError(dst)
        _single_copy(srcs[0], dst)


def remove(*files: str):
    """
    Overview:
        Remove a file or a directory at ``file``.
        ``file`` can be a file or a directory, both are supported.

    :param files: Files or directories to be removed.

    .. note::
        You can use this like ``rm -rf`` command on unix.

    Examples::
        >>> import os
        >>> from hbutils.system import remove
        >>>
        >>> os.listdir('.')
        ['test', 'README.md', 'cloc.sh', 'codecov.yml', 'new_test', 'new_path_2', 'setup.py', 'requirements-test.txt', 'new_path_1', 'requirements-doc.txt', 'requirements.txt', 'new_cloc.sh']
        >>>
        >>> remove('codecov.yml')  # remove file
        >>> remove('new_test')  # remove directory
        >>> os.listdir('.')
        ['test', 'README.md', 'cloc.sh', 'new_path_2', 'setup.py', 'requirements-test.txt', 'new_path_1', 'requirements-doc.txt', 'requirements.txt', 'new_cloc.sh']
        >>>
        >>> os.listdir('new_path_1')
        ['requirements-test.txt', 'requirements-doc.txt', 'requirements.txt']
        >>> remove('new_path_1/*.txt')  # remove files from directory
        >>> os.listdir('new_path_1')
        []
        >>>
        >>> print(*os.listdir('new_path_2'), sep=\'\\n\')
        test_version.py
        test_file.py
        test_type.py
        test_package.py
        test_implementation.py
        requirements-test.txt
        __init__.py
        test_directory.py
        requirements-doc.txt
        requirements.txt
        >>> remove('README.md', 'test/**/*.py', 'new_path_2/*.py')  # remove plenty of files
        >>> print(*os.listdir('new_path_2'), sep=\'\\n\')
        requirements-test.txt
        requirements-doc.txt
        requirements.txt
    """
    for file in glob(*files):
        try:  # remove directory
            shutil.rmtree(file)
        except NotADirectoryError:  # remove file
            os.remove(file)


def _single_getsize(file: str):
    if os.path.isfile(file):
        return os.path.getsize(file)
    else:
        total = 0
        for dirpath, dirnames, filenames in os.walk(file):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if not os.path.islink(fp):
                    total += os.path.getsize(fp)

        return total


def getsize(*files: str):
    """
    Overview:
        Get size of a file or a directory.

    :param files: File paths.
    :return: Size of the file or the total size of the directory.

    .. note::
        You can use this like ``du -sh`` command on unix.

    Examples::
        >>> from hbutils.system import getsize
        >>>
        >>> getsize('README.md')  # a file
        5368
        >>> getsize('test')  # a directory
        1575574
        >>> getsize('hbutils/**/*.py')  # glob filtered files
        238080
    """
    return sum(map(_single_getsize, glob(*files)))



"""
Overview:
    Reflection module, include some useful utilities for the python language.
"""
from .clazz import *
from .context import *
from .exception import *
from .func import *
from .imports import *
from .iter import *
from .module import *



"""
Overview:
    Isolation for stdin stream.
"""
import io
import os
from contextlib import _RedirectStream, contextmanager
from typing import List, Union, ContextManager, TextIO

from ...system import TemporaryDirectory

__all__ = [
    'isolated_stdin',
]


# noinspection PyPep8Naming
class _redirect_stdin(_RedirectStream):
    _stream = 'stdin'


def _to_input_text(v: Union[str, List[str]]) -> str:
    if isinstance(v, str):
        return v
    elif isinstance(v, list):
        return '\n'.join(v)
    else:
        raise TypeError(f'Unknown stdin type - {v!r}.')


@contextmanager
def _stdin_via_mem(v: str) -> ContextManager[TextIO]:
    with io.StringIO(v) as f:
        yield f


@contextmanager
def _stdin_via_file(v: str) -> ContextManager[TextIO]:
    with TemporaryDirectory() as tdir:
        stdin_file = os.path.join(tdir, 'stdin')
        with open(stdin_file, 'w+') as inf:
            inf.write(v)

        with open(stdin_file, 'r') as f:
            yield f


@contextmanager
def isolated_stdin(v: Union[str, List[str]], mem: bool = False):
    """
    Overview:
        Isolation for stdin stream.

    :param v: Input content, a whole string or a list of string supported.
    :param mem: Use memory or not. Default is ``False`` which means \
        a temporary file will be used as fake input stream.

    Examples::
        >>> from hbutils.testing import isolated_stdin
        >>> with isolated_stdin(['123', '456']):
        ...     a = int(input())
        ...     b = int(input())
        ...     print(a, b, a + b)
        123 456 579
    """
    context = _stdin_via_mem if mem else _stdin_via_file
    with context(_to_input_text(v)) as f:
        with _redirect_stdin(f):
            yield



from collections.abc import Mapping
from typing import Iterator, TypeVar

__all__ = [
    'StackedMapping',
]

_KeyType = TypeVar('_KeyType')
_ValueType = TypeVar('_ValueType')


class StackedMapping(Mapping):
    """
    Overview:
        Stacked mapping data structure.

        Multiple mapping-liked object (such as ``dict``) can be stacked together and accessed like one mapping object.

    .. note::
        :class:`StackedMapping` is readonly. The ``__setitem__`` and ``__delitem__``'s behaviour cannot be defined, \
        because which dict to write or delete can not be determined.

    Examples::
        >>> from hbutils.collection import StackedMapping
        >>>
        >>> d1 = {'a': 1, 'b': 2}
        >>> d2 = {'b': 3, 'c': 4, 'd': 5}
        >>> d3 = {'c': 6, 'e': 7}
        >>> s = StackedMapping(d1, d2, d3)  ## stack together, d3 > d2 > d1
        >>>
        >>> for key, value in s.items():  # __iter__
        ...     print(key, value)
        a 1
        b 3
        c 6
        d 5
        e 7
        >>> s['a']  # __getitem__, from d1['a']
        1
        >>> s['b']  # from d2['b'], d2 > d1
        3
        >>> s['c']  # from d3['c'], d3 > d2
        6
        >>> s['d']  # from d2['d']
        5
        >>> s['e']  # from d3['e']
        7
        >>> 'c' in s  # __contains__
        True
        >>> 'f' in s  # 'f' not found in neither d1, d2 nor d3
        False
        >>> s == {'a': 1, 'b': 3, 'c': 6, 'd': 5, 'e': 7}  # __iter__
        True
        >>> len(s)  # __len__
        5
        >>>
        >>> d2['c'] = 11  ## update original dicts
        >>> del d2['b']
        >>> del d3['c']
        >>> del d1['b']
        >>>
        >>> s['a']  # __getitem__, from d1['a']
        1
        >>> s['b']  # 'b' nor found in neither d1, d2 nor d3
        KeyError: 'b'
        >>> s['c']  # from d2['c']
        11
        >>> s['d']  # from d2['d']
        5
        >>> s['e']  # from d3['e']
        7
        >>> len(s)  # 'b' is no longer here
        4
    """

    def __init__(self, *mps: Mapping):
        """
        Constructor of :class:`StackedMapping`.

        Later mappings will be stacked on top of the earlier ones.

        :param mps: Multiple mapping objects.
        """
        self._mps = mps

    def __getitem__(self, k: _KeyType) -> _ValueType:
        for m in reversed(self._mps):
            try:
                return m[k]
            except KeyError:
                continue

        raise KeyError(k)

    def _key_iter(self):
        _exist_keys = set()
        for m in self._mps:
            for k in m.keys():
                if k not in _exist_keys:
                    _exist_keys.add(k)
                    yield k

    def __len__(self) -> int:
        return len(list(self._key_iter()))

    def __iter__(self) -> Iterator[_KeyType]:
        yield from self._key_iter()



"""
Overview:
    Random state and seed management.

    Native default random instance (``random._inst``), torch, numpy and faker (default random instance) are supported. \
    Other random sources can be registered with :func:`register_random_source` and :func:`register_random_instance`.
"""
import random
import warnings
from contextlib import contextmanager
from functools import wraps
from typing import Callable, Tuple, TypeVar, Dict, Mapping

T = TypeVar('T')

_SEED_FUNC = Callable[[int], None]
_GETSTATE_FUNC = Callable[[], T]
_SETSTATE_FUNC = Callable[[T], None]
_RANDOM_SOURCES: Dict[str, Tuple[_SEED_FUNC, _GETSTATE_FUNC, _SETSTATE_FUNC]] = {}

__all__ = [
    'register_random_source',
    'register_random_instance',
    'get_global_state',
    'set_global_state',
    'keep_global_state',
    'global_seed',
    'seedable_func',
]


def register_random_source(name: str, seed: _SEED_FUNC, getstate: _GETSTATE_FUNC, setstate: _SETSTATE_FUNC):
    """
    Overview:
        Register random source, providing name, random seed function, state getting function \
        and state setting function.

    :param name: Name of random source.
    :param seed: Seed function, format: ``seed(x)``.
    :param getstate: State getting function, format: ``getstate()``.
    :param setstate: State setting function, format: ``setstate(state)``.

    Examples::
        >>> import random
        >>> from hbutils.random import global_seed, register_random_source
        >>>
        >>> rnd = random.Random()  # custom random object
        >>> global_seed(0)  # try to use same seed
        >>> rnd.random()
        0.4563765178328746
        >>> global_seed(0)
        >>> rnd.random()  # not same
        0.06875325446897462
        >>>
        >>> register_random_source('custom_random', rnd.seed, rnd.getstate, rnd.setstate)
        >>> global_seed(0)  # try again
        >>> rnd.random()
        0.8444218515250481
        >>> global_seed(0)
        >>> rnd.random()  # the same
        0.8444218515250481

    """
    if name in _RANDOM_SOURCES:
        raise NameError(f'Name {name!r} already exist.')
    _RANDOM_SOURCES[name] = (seed, getstate, setstate)


def register_random_instance(name: str, rnd: random.Random):
    """
    Overview:
        Register custom random instance.

    :param name: Name of random source.
    :param rnd: Custom random instance, should be an instance of :class:`random.Random`.

    Examples::
        >>> import random
        >>> from hbutils.random import global_seed, register_random_instance
        >>>
        >>> rnd = random.Random()  # custom random object
        >>> global_seed(0)  # try to use same seed
        >>> rnd.random()
        0.48936053503964005
        >>> global_seed(0)
        >>> rnd.random()  # not same
        0.4113361070387721
        >>>
        >>> register_random_instance('custom_random', rnd)
        >>> global_seed(0)  # try again
        >>> rnd.random()
        0.8444218515250481
        >>> global_seed(0)
        >>> rnd.random()  # the same
        0.8444218515250481
    """
    register_random_source(name, rnd.seed, rnd.getstate, rnd.setstate)


def get_global_state() -> Dict[str, T]:
    """
    Overview:
        Get states of registered random sources.

    :returns: Global states.

    Examples::
        >>> import random
        >>> import numpy as np
        >>> import torch
        >>> from faker import Faker
        >>>
        >>> from hbutils.random import get_global_state, set_global_state
        >>>
        >>> _ = random.randint(0, 100)  # just do something
        >>> _ = random.random()
        >>> _ = torch.randn(2, 3)
        >>> _ = np.random.randn(2, 3)
        >>>
        >>> states = get_global_state()
        >>> random.randint(0, 100)  # first time's result
        99
        >>> random.random()
        0.4656250864192085
        >>> torch.randn(2, 3)
        tensor([[ 0.8886, -0.3602,  1.3071],
                [-0.0187, -0.5980, -0.5469]])
        >>> np.random.randn(2, 3)
        array([[ 1.24249156,  0.71018699, -0.53496231],
               [ 0.78748336, -0.01407442, -0.6607438 ]])
        >>> Faker().sentence(5)
        'New pass crime most.'
        >>>
        >>> set_global_state(states)
        >>> random.randint(0, 100)  # same as the first time
        99
        >>> random.random()
        0.4656250864192085
        >>> torch.randn(2, 3)
        tensor([[ 0.8886, -0.3602,  1.3071],
                [-0.0187, -0.5980, -0.5469]])
        >>> np.random.randn(2, 3)
        array([[ 1.24249156,  0.71018699, -0.53496231],
               [ 0.78748336, -0.01407442, -0.6607438 ]])
        >>> Faker().sentence(5)
        'New pass crime most.'
    """
    return {name: getstate() for name, (_, getstate, _) in _RANDOM_SOURCES.items()}


def set_global_state(states: Mapping[str, T]):
    """
    Overview:
        Set states of registered random sources.

    :param states: States to be set.

    Examples::
        See :func:`get_global_state`.
    """
    _skipped_names = set()
    _existing_names = set(_RANDOM_SOURCES.keys())
    for name, state in states.items():
        if name not in _existing_names:
            _skipped_names.add(name)
        else:
            _, _, setstate = _RANDOM_SOURCES[name]
            setstate(state)
            _existing_names.remove(name)

    if _skipped_names:
        _skipped_names = tuple(sorted(_skipped_names))
        warnings.warn(f'Random source {_skipped_names} skipped due to their non-existence in this environment.')

    if _existing_names:
        _existing_names = tuple(sorted(_existing_names))
        warnings.warn(f'Random source {_existing_names} not recovered because they are not provided.')


@contextmanager
def keep_global_state():
    """
    Overview:
        Keep all the states during this period.

    Examples::
        >>> import torch
        >>> from hbutils.random import global_seed, keep_global_state
        >>>
        >>> global_seed(0)
        >>> torch.randn(2, 3)  # before value 1
        tensor([[ 1.5410, -0.2934, -2.1788],
                [ 0.5684, -1.0845, -1.3986]])
        >>> torch.randn(2, 3)  # after value 1
        tensor([[ 0.4033,  0.8380, -0.7193],
                [-0.4033, -0.5966,  0.1820]])
        >>>
        >>> global_seed(0)
        >>> torch.randn(2, 3)  # before value 2, same as 1
        tensor([[ 1.5410, -0.2934, -2.1788],
                [ 0.5684, -1.0845, -1.3986]])
        >>> with keep_global_state():  # do anything you want here
        ...     _ = torch.randn(100, 200, 2)
        ...     _ = torch.randint(20, 30, (30, 40))
        >>> torch.randn(2, 3)  # after value 2, same as 1
        tensor([[ 0.4033,  0.8380, -0.7193],
                [-0.4033, -0.5966,  0.1820]])
    """
    states = get_global_state()
    try:
        yield
    finally:
        set_global_state(states)


def global_seed(seed: int):
    """
    Overview:
        Set seed for all registered random source.

    :param seed: Random seed.

    Examples::
        See :func:`keep_global_state` and :func:`register_random_instance`.

    """
    for _, (fseed, _, _) in _RANDOM_SOURCES.items():
        fseed(seed)


def seedable_func(func):
    """
    Overview:
        Wrap function to support seed.

    :param func: Function to be decorated, a new ``seed`` argument will be added.

    Examples::
        >>> import torch
        >>> from hbutils.random import seedable_func
        >>>
        >>> @seedable_func
        ... def get_random_value(mean, std):
        ...     return torch.randn((2, 3)) * std + mean
        >>>
        >>> get_random_value(2, 3)
        tensor([[-0.0844,  5.2530,  3.4248],
                [ 4.4923,  0.0492,  3.6731]])
        >>> get_random_value(2, 3)  # not the same
        tensor([[2.7600, 2.5135, 0.6484],
                [1.2459, 0.1020, 2.5905]])
        >>>
        >>> get_random_value(2, 3, seed=0)
        tensor([[ 6.6230,  1.1197, -4.5364],
                [ 3.7053, -1.2536, -2.1958]])
        >>> get_random_value(2, 3, seed=1)
        tensor([[3.9841, 2.8008, 2.1850],
                [3.8640, 0.6443, 1.5016]])
        >>> get_random_value(2, 3, seed=0)  # repeatable
        tensor([[ 6.6230,  1.1197, -4.5364],
                [ 3.7053, -1.2536, -2.1958]])
    """

    @wraps(func)
    def _new_func(*args, seed: int = None, **kwargs):
        if seed is not None:
            global_seed(seed)

        return func(*args, **kwargs)

    return _new_func


register_random_source('native_random', random.seed, random.getstate, random.setstate)

try:
    import numpy as np
except ImportError:
    pass
else:
    register_random_source('numpy', np.random.seed, np.random.get_state, np.random.set_state)

try:
    import torch
except ImportError:
    pass
else:
    register_random_source('torch', torch.manual_seed, torch.get_rng_state, torch.set_rng_state)

try:
    from faker import Faker
except ImportError:
    pass
else:
    _FAKER = Faker()
    register_random_instance('faker_default', _FAKER.random)



import random
from typing import Iterator, Mapping, Optional, List, Tuple, Union

from hbutils.model import visual, hasheq, accessor, asitems, IComparable
from .base import BaseGenerator
from ...reflection import nested_for, progressive_for

__all__ = [
    'AETGGenerator',
]


@hasheq()
@visual()
@accessor(readonly=True)
@asitems(['name', 'value'])
class _NameValueTuple(IComparable):
    def __init__(self, name, value):
        self.__name = name
        self.__value = value

    def _cmpkey(self):
        return self.__name, self.__value


@hasheq()
@visual()
@accessor(readonly=True)
@asitems(['items'])
class _AETGValuePair(IComparable):
    def __init__(self, *pairs):
        self.__items = tuple(sorted(pairs, key=lambda x: (x.name, x.value)))

    def _cmpkey(self):
        return self.__items


def _create_init_pairs(names: List[str]) -> List[Tuple[str, ...]]:
    return list(progressive_for(names, min(2, len(names))))


def _process_pairs(pairs: List[Tuple[str, ...]], names: List[str]) -> List[Tuple[str, ...]]:
    _name_set = set(names)
    _name_id_dict = {name: i for i, name in enumerate(names)}

    _init_pairs = []
    for pair in pairs:
        actual_pair_ids = sorted(set(map(lambda x: _name_id_dict[x], pair)))
        actual_pair = tuple(names[i] for i in actual_pair_ids)
        _init_pairs.append(actual_pair)

    _init_pairs = sorted(_init_pairs, key=lambda x: (len(x), x), reverse=True)
    _final_pairs = []
    for pair in _init_pairs:
        sp = set(pair)
        is_included = False
        for exist_pair in _final_pairs:
            sep = set(exist_pair)
            if sp & sep == sp:
                is_included = True
                break

        if not is_included:
            _final_pairs.append(pair)

    return _final_pairs[::-1]


# noinspection PyProtectedMember
_DEFAULT_RANDOM = random._inst


def _to_random(rnd: Optional[Union[random.Random, int]] = None):
    if isinstance(rnd, random.Random):
        return rnd
    elif isinstance(rnd, int):
        return random.Random(rnd)
    elif rnd is None:
        return _DEFAULT_RANDOM
    else:
        raise TypeError(f'Unknown random type - {repr(rnd)}.')


class AETGGenerator(BaseGenerator):
    """
    Full AETG model, test cases will be generated to make sure the required pairs will be all tested.
    """

    def __init__(self, values, names: Optional[List[str]] = None,
                 pairs: Optional[List[Tuple[str, ...]]] = None,
                 rnd: Optional[Union[random.Random, int]] = None):
        """
        Constructor of the :class:`hbutils.testing.AETGGenerator` class.

        :param values: Selection values, such as ``{'a': [2, 3], 'b': ['b', 'c']}``.
        :param names: Names of the given generator, default is ``None`` which means use the sorted \
            key set of the values.
        :param pairs: Pairs required to be all tested, default is ``None`` which means all the \
            binary pairs will be included.
        :param rnd: Random object or int-formatted random seed, \
            default is ``None`` which means the default random object will be used.
        """
        BaseGenerator.__init__(self, values, names)
        if pairs is None:
            pairs = _create_init_pairs(self.names)
        self.__pairs = _process_pairs(pairs, self.names)
        self.__rnd = _to_random(rnd)

        self.__node_cnt = None
        self.__non_exist_pairs = None

    @property
    def pairs(self) -> List[Tuple[str, ...]]:
        """
        Pairs required to be all tested.
        """
        return self.__pairs

    def __get_init_info(self) -> Tuple[dict, list, set]:
        if self.__node_cnt is None:
            node_cnt = {}
            non_exist_pairs = set()
            for one_pair in self.__pairs:
                for value_items in nested_for(*[self.values[name] for name in one_pair]):
                    pair_items = []
                    for name, value in zip(one_pair, value_items):
                        tp = _NameValueTuple(name, value)
                        pair_items.append(tp)
                        node_cnt[tp] = node_cnt.get(tp, 0) + 1
                    non_exist_pairs.add(_AETGValuePair(*pair_items))

            self.__node_cnt = node_cnt
            self.__non_exist_pairs = sorted(non_exist_pairs)

        return dict(self.__node_cnt), list(self.__non_exist_pairs), set(self.__non_exist_pairs)

    def cases(self) -> Iterator[Mapping[str, object]]:
        """
        Get the cases in this AETG model.

        Examples::
            >>> from hbutils.testing import AETGGenerator
            >>> gene = AETGGenerator({'a': (1, 2), 'b': (3, 4), 'c': (5, 6), 'd': (7, 8), 'e': (9, 10)})
            >>> for p in gene.cases():
            ...     print(p)
            {'a': 1, 'b': 3, 'c': 6, 'd': 8, 'e': 10}
            {'a': 2, 'b': 4, 'c': 5, 'd': 7, 'e': 9}
            {'a': 2, 'b': 3, 'c': 6, 'd': 7, 'e': 9}
            {'a': 2, 'b': 4, 'c': 6, 'd': 8, 'e': 10}
            {'a': 1, 'b': 3, 'c': 5, 'd': 7, 'e': 10}
            {'a': 1, 'b': 4, 'c': 5, 'd': 8, 'e': 9}
            >>> gene = AETGGenerator(
            ...     {'a': (1, 2), 'b': (3, 4), 'c': (5, 6), 'd': (7, 8), 'e': (9, 10)},
            ...     pairs=[('a', 'c'), ('b', 'd'), ('e',)]
            ... )
            >>> for p in gene.cases():
            ...     print(p)
            {'a': 2, 'b': 3, 'c': 6, 'd': 8, 'e': 9}
            {'a': 1, 'b': 4, 'c': 5, 'd': 7, 'e': 10}
            {'a': 2, 'b': 4, 'c': 5, 'd': 8, 'e': 9}
            {'a': 1, 'b': 3, 'c': 6, 'd': 7, 'e': 10}
        """
        n = len(self.names)
        m = len(self.__pairs[-1]) if self.__pairs else 0
        node_cnt, non_exist_pairs_list, non_exist_pairs_set = self.__get_init_info()

        while non_exist_pairs_set:
            first_pair = None
            while non_exist_pairs_list:
                _pair = non_exist_pairs_list.pop()
                if _pair in non_exist_pairs_set:
                    non_exist_pairs_list.append(_pair)
                    first_pair = _pair
                    break

            tnames = []
            seqs = []
            for pair_item in first_pair.items:
                tnames.append(pair_item.name)
                seqs.append(pair_item)

            _tname_set = set(tnames)
            other_names = [name for name in self.names if name not in _tname_set]
            self.__rnd.shuffle(other_names)
            tnames += other_names

            for i in range(len(seqs), n):
                iname = tnames[i]
                curpair, curxk = None, None
                for ivalue in self.values[iname]:
                    ituple = _NameValueTuple(iname, ivalue)

                    litems = [ituple]
                    non_exists = 0
                    for j in range(0, min(m, i + 1)):
                        if j > 0:
                            litems.append(seqs[i - j])
                        now_pair = _AETGValuePair(*litems)
                        if now_pair in non_exist_pairs_set:
                            non_exists += 1

                    xk = (non_exists, node_cnt.get(ituple, 0), self.__rnd.random())
                    if curxk is None or xk > curxk:
                        curpair = ituple
                        curxk = xk

                seqs.append(curpair)

            px = {pair.name: pair.value for pair in seqs}
            for one_pair in self.__pairs:
                new_pair = _AETGValuePair(*(_NameValueTuple(name, px[name]) for name in one_pair))
                if new_pair in non_exist_pairs_set:
                    non_exist_pairs_set.remove(new_pair)
                    for np in new_pair.items:
                        node_cnt[np] -= 1

            yield {name: px[name] for name in self.names}



"""
Overview:
    String process module, include some useful utilities for string.
"""
from .inflection import *
from .plural import *
from .template import *
from .tree import *
from .trunc import *



import logging
from contextlib import contextmanager
from typing import Optional, Union, List

from ...collection.recover import get_recovery_func

__all__ = [
    'isolated_logger',
]


@contextmanager
def isolated_logger(logger: Optional[Union[str, logging.Logger]] = None,
                    handlers: Optional[List[logging.Handler]] = None,
                    close_handlers: bool = True):
    """
    Overview:
        Mock loggers in `logging` module.

    :param logger: Logger or logger's name for isolation.
    :param handlers: Initial handlers for isolation.
    :param close_handlers: Close handler's after complete.

    Examples::
        >>> import logging
        >>> from rich.logging import RichHandler  # a 3rd-party logger
        >>> from hbutils.testing import isolated_logger
        >>>
        >>> logging.error('this is error')  # normal log
        ERROR:root:this is error
        >>> logging.error('this is error')
        ERROR:root:this is error

        >>> with isolated_logger(handlers=[  # replaced with custom handlers
        ...     RichHandler(),
        ...     logging.FileHandler('test_log.txt'),
        ... ]):
        ...     logging.error('this is error inside 1')
        ...     logging.error('this is error inside 2')
        [11/08/22 15:39:37] ERROR    this is error inside 1                    <stdin>:5
                            ERROR    this is error inside 2                    <stdin>:6

        >>> logging.error('this is error')  # change back to normal log
        ERROR:root:this is error
        >>> logging.error('this is error')
        ERROR:root:this is error

    """
    if not isinstance(logger, logging.Logger):
        logger = logging.getLogger(logger)
    handlers = list(handlers or [])

    rf = get_recovery_func(logger)
    try:
        logger.handlers[:] = handlers[:]
        yield logger
    finally:
        if close_handlers:
            for handler in logger.handlers:
                handler.close()
        rf()



from .clazz import asitems, visual, hasheq, accessor
from ..reflection import dynamic_call, frename

__all__ = [
    'raw_support'
]


def raw_support(condition, raw_name: str = 'raw', unraw_name: str = 'unraw', proxy_name='RawProxy'):
    """
    Overview:
        Get raw support wrapper function and class.

    Arguments:
        - condition: Condition of wrapper, the value will be wrapped only when ``condition`` is satisfied.
        - raw_name (:obj:`str`): Raw function's name.
        - unraw_name (:obj:`str`): Unraw function's name.
        - proxy_name (:obj:`str`): Proxy class's name.

    Returns:
        - (raw_func, unraw_func, proxy_class).

    Examples::
        >>> from hbutils.model import raw_support
        >>> raw, unraw, RawProxy = raw_support(lambda x: isinstance(x, dict))
        >>> raw(1)
        1
        >>> raw({'a': 1, 'b': 2})
        <RawProxy value: {'a': 1, 'b': 2}>

        >>> unraw(1)
        1
        >>> unraw({'a': 1, 'b': 2})
        {'a': 1, 'b': 2}
        >>> unraw(raw({'a': 1, 'b': 2}))
        {'a': 1, 'b': 2}
    """
    _dc = dynamic_call(condition)

    @visual()
    @hasheq()
    @accessor(readonly=True)
    @asitems(['value'])
    class _RawProxy:

        def __init__(self, value):
            self.__value = value

    _RawProxy.__name__ = proxy_name

    @frename(raw_name)
    def _raw(v):
        if not isinstance(v, _RawProxy) and _dc(v):
            return _RawProxy(v)
        else:
            return v

    @frename(unraw_name)
    def _unraw(v):
        if isinstance(v, _RawProxy):
            return v.value
        else:
            return v

    return _raw, _unraw, _RawProxy



import io
import os
import textwrap
from typing import List, Union

__all__ = [
    'TextAligner',
]

from ...design import SingletonMark

from ...string import plural_word

_NOT_GIVEN = SingletonMark('_TextAligner_NOT_GIVEN')


class TextAligner:
    """
    Overview:
        Text aligner for comparing texts in unittest.
    """

    def __init__(self, line_rstrip: bool = True, keep_empty_tail: bool = False,
                 text_func=None, line_func=None, ls_func=None):
        """
        Constructor of :class:`TextAligner`.

        :param line_rstrip: Right strip each line, default is ``True``.
        :param keep_empty_tail: Keep the empty tail of the text, default is ``False``, which means the \
            empty tails will be ignored.
        :param text_func: Function for preprocessing whole text.
        :param line_func: Function for processing each line.
        :param ls_func: Function for processing lines list.
        """
        self.__line_rstrip = line_rstrip
        self.__keep_empty_tail = keep_empty_tail
        self.__text_func = text_func or (lambda x: x)
        self.__line_func = line_func or (lambda x: x)
        self.__ls_func = ls_func or (lambda x: x)

    def text_trans(self, text_func):
        """
        Transformation for the original text.

        :param text_func: New text process function.
        :return: A new :class:`TextAligner` object with ``text_func`` process.
        """
        return self.__class__(
            self.__line_func, self.__keep_empty_tail,
            lambda x: text_func(self.__text_func(x)),
            self.__line_func, self.__ls_func
        )

    def line_map(self, line_func):
        """
        Mapping for the text of each line.

        :param line_func: New line process function.
        :return: A new :class:`TextAligner` object with ``line_func`` process.
        """
        return self.__class__(
            self.__line_func, self.__keep_empty_tail,
            self.__text_func,
            lambda x: line_func(self.__line_func(x)),
            self.__ls_func,
        )

    def ls_trans(self, ls_func):
        """
        Transformation for the separated lines.

        :param ls_func: New lines process function.
        :return: A new :class:`TextAligner` object with ``ls_func`` process.
        """
        return self.__class__(
            self.__line_func, self.__keep_empty_tail,
            self.__text_func, self.__line_func,
            lambda x: ls_func(self.__ls_func(x)),
        )

    def __getattr__(self, item: str) -> '_StrMethodProxy':
        """
        Append postprocess from :class:`str` of each line.

        :param item: Method name.

        Examples::
            >>> from hbutils.testing import TextAligner
            >>> text_align = TextAligner()
            >>> print(text_align.lower().multiple_lines()('''
            ... Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... '''))
            python 3.6.5
            hello world!!
              do not see me like this...
        """
        return _StrMethodProxy(self, item)

    def multiple_lines(self, lstrip: bool = True, dedent: bool = True):
        """
        Switch to multiple-line mode.

        :param lstrip: Left strip the original text.
        :param dedent: Dedent the text.

        Examples::
            >>> from hbutils.testing import TextAligner
            >>> text_align = TextAligner()
            >>> print(text_align.multiple_lines()('''
            ... Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... '''))
            Python 3.6.5
            Hello world!!
              Do not see me like this...

        .. note::
            With :meth:`multiple_lines`, the text's comparison will be compatible with text wrapper.
        """
        align = self
        if dedent:
            align = align.text_trans(textwrap.dedent)
        if lstrip:
            align = align.text_trans(str.lstrip)

        return align

    def _process(self, text: Union[str, List[str]]) -> List[str]:
        if isinstance(text, (list, tuple)):
            return self._process(os.linesep.join(text))
        elif isinstance(text, str):
            if not self.__keep_empty_tail:
                text = text.rstrip()
            text = self.__text_func(text)

            lines = text.splitlines(keepends=False)
            if self.__line_rstrip:
                lines = list(map(str.rstrip, lines))
            lines = list(map(self.__line_func, lines))
            lines = list(self.__ls_func(lines))

            return lines
        else:
            raise TypeError(f'Invalid content type - {text!r}.')

    def splitlines(self, text: Union[str, List[str]]) -> List[str]:
        """
        Transform the original text or lines to list of aligned lines.

        :param text: Original text or lines.
        :return: List of split lines.

        Examples::
            >>> from hbutils.testing import TextAligner
            >>> text_align = TextAligner()
            >>> print(text_align.splitlines('''Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... '''))
            ['Python 3.6.5', 'Hello world!!', '  Do not see me like this...']

            >>> print(text_align.lstrip().splitlines('''Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... '''))
            ['Python 3.6.5', 'Hello world!!', 'Do not see me like this...']
        """
        return self._process(text)

    def __call__(self, text: Union[str, List[str]]) -> str:
        """
        Transform the original text or lines to aligned text.

        :param text: Original text or lines.
        :return: Aligned text.

        Examples::
            >>> from hbutils.testing import TextAligner
            >>> text_align = TextAligner()
            >>> print(text_align('''Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... '''))
            Python 3.6.5
            Hello world!!
              Do not see me like this...

            >>> print(text_align.lstrip()('''Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... '''))
            Python 3.6.5
            Hello world!!
            Do not see me like this...
        """
        return os.linesep.join(self._process(text))

    @staticmethod
    def _eq_compare_message(expect: List[str], actual: List[str], max_diff: int = 3, max_extra: int = 5):
        if expect == actual:  # pragma: no cover
            return 'No difference found.'
        else:
            with io.StringIO() as sf:
                if len(expect) != len(actual):
                    print(f'{plural_word(len(expect), "line")} expected, '
                          f'but {plural_word(len(actual), "line")} found actually.', file=sf)

                diff_cnt = 0
                for lineno, (eline, aline) in enumerate(zip(expect, actual), start=1):
                    if eline != aline:
                        diff_cnt += 1
                        if max_diff <= 0 or diff_cnt <= max_diff:
                            print(f'Difference found in line {lineno}:', file=sf)
                            print(f'    Expect: {eline}', file=sf)
                            print(f'    Actual: {aline}', file=sf)
                if 0 < max_diff < diff_cnt:
                    print(f'    ... ({plural_word(diff_cnt - max_diff, "more different line")}) ...', file=sf)

                if len(expect) != len(actual):
                    common_length = min(len(expect), len(actual))
                    if len(expect) > common_length:
                        print(f'Another {plural_word(len(expect) - common_length, "extra line")} '
                              f'found in expected lines:', file=sf)
                        extra_lines = expect[common_length:]
                    elif len(actual) > common_length:
                        print(f'Another {plural_word(len(actual) - common_length, "extra line")} '
                              f'found in actual lines:', file=sf)
                        extra_lines = actual[common_length:]

                    if max_extra > 0:
                        elines = extra_lines[:max_extra]
                    else:
                        elines = extra_lines
                    for line in elines:
                        print(f'    | {line}', file=sf)

                    if len(elines) < len(extra_lines):
                        print(f'    ... ({plural_word(len(extra_lines) - len(elines), "more line")}) ...', file=sf)

                return sf.getvalue()

    def assert_equal(self, expect: Union[str, List[str]], actual: Union[str, List[str]],
                     max_diff: int = 3, max_extra: int = 5):
        """
        Assert two string is equal.

        :param expect: Expected text or lines.
        :param actual: Actual text or lines.
        :param max_diff: Max different lines to show in assertion message, default is ``3``.
        :param max_extra: Max extra lines to show in assertion message, default is ``5``.

        Examples::
            >>> from hbutils.testing import TextAligner
            >>> text_align = TextAligner()
            >>> text_align.multiple_lines().assert_equal('''Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... ''', '''
            ...         Python 3.6.5
            ...         Hello world!!
            ...           Do not see me like this...
            ... ''')  # this is okay

            >>> text_align.multiple_lines().assert_equal('''Python 3.6.5
            ... Hello world!!
            ...   Do not see me like this...
            ... \\n\\n\\t\\n
            ... ''', '''
            ...         Python 3.6.5
            ...         Hello world!!
            ...         Do not see me like this...
            ... ''')
            AssertionError: Difference found in line 3:
                Expect:   Do not see me like this...
                Actual: Do not see me like this...
        """
        expect, actual = self._process(expect), self._process(actual)
        assert expect == actual, self._eq_compare_message(expect, actual, max_diff, max_extra)

    @staticmethod
    def _ne_compare_message(expect: List[str], actual: List[str]):
        if expect != actual:  # pragma: no cover
            return 'Difference found in actual text.'
        else:
            return 'Actual text are completely the same as expected one!'

    def assert_not_equal(self, expect: Union[str, List[str]], actual: Union[str, List[str]]):
        """
        Assert two string is not equal, which is similar to :meth:`assert_equal`.

        :param expect: Expected text or lines.
        :param actual: Actual text or lines.
        """
        expect, actual = self._process(expect), self._process(actual)
        assert expect != actual, self._ne_compare_message(expect, actual)


class _StrMethodProxy:
    def __init__(self, align: TextAligner, name: str):
        self.__align = align
        if hasattr(str, name) and callable(getattr(str, name)):
            self.__func = getattr(str, name)
        else:
            raise AttributeError(f'Attribute {name!r} not found in str.')

    def __call__(self, *args, **kwargs) -> 'TextAligner':
        return self.__align.line_map(lambda x: self.__func(x, *args, **kwargs))



from ...system import which

__all__ = ['cmdv']


def cmdv(execfile: str) -> bool:
    """
    Overview:
        Check if the given command is exist in this environment, like the ``command -v xxx`` command in Linux.

    :param execfile: Executable file, such as ``python``.
    :return: This executable file is exist or not.

    Examples::
        >>> from hbutils.testing import cmdv
        >>>
        >>> cmdv('apt-get')  # should exist on Ubuntu
        True
        >>> cmdv('bash')  # should exist on Linux
        True
        >>> cmdv('cmd')  # should exist on Windows
        False
        >>> cmdv('not_installed')
        False
    """
    return bool(which(execfile))



"""
Overview:
    Meta information for hbutils package.
"""

#: Title of this project (should be `hbutils`).
__TITLE__ = "hbutils"

#: Version of this project.
__VERSION__ = "0.9.0"

#: Short description of the project, will be included in ``setup.py``.
__DESCRIPTION__ = 'Some useful functions and classes in Python infrastructure development.'

#: Author of this project.
__AUTHOR__ = "HansBug"

#: Email of the authors'.
__AUTHOR_EMAIL__ = "hansbug@buaa.edu.cn"



from typing import Union, TypeVar, Sequence, Callable, Optional, Dict, List, Iterable

__all__ = [
    'unique',
    'group_by',
]

_ElementType = TypeVar('_ElementType')


def unique(s: Union[Sequence[_ElementType]]) -> Sequence[_ElementType]:
    """
    Overview:
        Unique all the values in the given ``s``, preserving its original order.

    :param s: Original sequence.
    :return: Unique sequence, with the original type.

    Examples::
        >>> from hbutils.collection import unique
        >>>
        >>> unique([1, 2, 3, 1])
        [1, 2, 3]
        >>> unique(('a', 'b', 'a', 'c', 'd', 'e', 'b'))
        ('a', 'b', 'c', 'd', 'e')
        >>> unique([3, 1, 2, 1, 4, 3])
        [3, 1, 2, 4]
    """
    _set, _result = set(), []
    for element in s:
        if element not in _set:
            _result.append(element)
            _set.add(element)

    return type(s)(_result)


_GroupType = TypeVar('_GroupType')
_ResultType = TypeVar('_ResultType')


def group_by(s: Iterable[_ElementType],
             key: Callable[[_ElementType], _GroupType],
             gfunc: Optional[Callable[[List[_ElementType]], _ResultType]] = None) -> Dict[_GroupType, _ResultType]:
    """
    Overview:
        Divide the elements into groups.

    :param s: Elements.
    :param key: Group key, should be a callable object.
    :param gfunc: Post-process function for groups, should be a callable object. Default is ``None`` which means \
        no post-processing will be performed.
    :return: Grouping result.

    Examples::
        >>> from hbutils.collection import group_by
        >>>
        >>> foods = [
        ...     'apple', 'orange', 'pear',
        ...     'banana', 'fish', 'pork', 'milk',
        ... ]
        >>> group_by(foods, len)  # group by length
        {5: ['apple'], 6: ['orange', 'banana'], 4: ['pear', 'fish', 'pork', 'milk']}
        >>> group_by(foods, len, len)  # group and get length
        {5: 1, 6: 2, 4: 4}
        >>> group_by(foods, lambda x: x[0])  # group by first letter
        {'a': ['apple'], 'o': ['orange'], 'p': ['pear', 'pork'], 'b': ['banana'], 'f': ['fish'], 'm': ['milk']}
        >>> group_by(foods, lambda x: x[0], len)  # group and get length
        {'a': 1, 'o': 1, 'p': 2, 'b': 1, 'f': 1, 'm': 1}
    """

    gfunc = gfunc or (lambda x: x)

    _result_dict: Dict[_GroupType, List[_ElementType]] = {}
    for item in s:
        _item_key = key(item)
        if _item_key not in _result_dict:
            _result_dict[_item_key] = []
        _result_dict[_item_key].append(item)

    return {
        key: gfunc(grps)
        for key, grps in _result_dict.items()
    }
